/**
 * \file cuesplitter.cpp
 * Splitter of audio files by cue file
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date Apr 15 2013
 *
 * Copyright (C) 2013  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cuesplitter.h"

#include "qoobarglobals.h"
#include "enums.h"
#include "application.h"
#include "tagger.h"
#include "tagsreaderwriter.h"

#include <QFileInfo>
#include <QTextStream>
#include <QFileSystemWatcher>
#include <QProcessEnvironment>
#include <QProcess>
#include <QTextCodec>
#include <QDir>

CueSplitter::CueSplitter(QObject *parent) :
    QObject(parent), process(0)
{
#ifdef Q_OS_WIN
    if (!programInstalled(QSL("shntool"),0) || !programInstalled(QSL("flac"),0))
        emit message(MT_WARNING,tr("Make sure you have shntool.exe and flac.exe "
                             "copied into the folder with qoobar.exe.<br>"
                             "You can download them from"
                             " <a href=\"http://etree.org/shnutils/shntool/\">"
                             "http://etree.org/shnutils/shntool/</a>"));
#else
    QStringList warnings;
    if (!programInstalled(QSL("shntool"),0)) warnings << tr("Please install shntool.");
    if (!programInstalled(QSL("flac"),0)) warnings << tr("Please install flac.");
    emit message(MT_WARNING, warnings.join(QSL(" ")));
#endif


}

CueSplitter::~CueSplitter()
{
    if (process) {
        process->kill();
        process->deleteLater();
    }
}

void CueSplitter::stop()
{
    if (process) process->kill();
}

void CueSplitter::split()
{
    //1. searching input file in the cue file
    QFile file(_cueFile);
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
        emit message(MT_ERROR, tr("Cannot open cue file %1").arg(_cueFile));
        return;
    }
    QTextStream in(&file);
    in.setCodec(QTextCodec::codecForLocale());
    in.setAutoDetectUnicode(true);
    QString inputFile;
    int trackCount=0;
    QRegExp r(QSL("FILE\\s+\"?([^\"]+)\"?"));
    while(!in.atEnd()) {
        QString s = in.readLine();
        if (r.indexIn(s)>-1) {
            inputFile=r.cap(1).trimmed();
            //break;
        }
        s=s.simplified();
        if (s.startsWith(QLS("TRACK ")))
            trackCount = s.section(QSL(" "),1,1).toInt();
    }
    file.close();
    emit trackCountChanged(trackCount);

    //2. splitting
    if (QFileInfo(inputFile).isRelative())
        inputFile = QString("%1/%2").arg(_outputDir).arg(inputFile);
    if (inputFile.isEmpty() || !QFileInfo(inputFile).exists()) {
        emit message(MT_ERROR, tr("Cannot find the file to split: %1").arg(inputFile));
        return;
    }

    //check the input file type and show warnings if needed
    QString inputFileExt = QFileInfo(inputFile).suffix();
    if (inputFileExt.toLower()=="ape" && !programInstalled(QSL("mac"),0)) {
#ifdef Q_OS_WIN
        emit message(MT_ERROR, tr("Make sure you have mac.exe "
                             "copied into the folder with qoobar.exe.<br>"
                             "You can download it from"
                             " <a href=\"http://etree.org/shnutils/shntool/\">"
                             "http://etree.org/shnutils/shntool/</a>"));
#endif
#ifdef Q_OS_LINUX
        emit message(MT_ERROR, tr("Please install mac. You can download it from "
                             "<a href=http://members.iinet.net.au/~aidanjm/mac-3.99-u4_b3-1_i386.deb>"
                             "http://members.iinet.net.au/~aidanjm/mac-3.99-u4_b3-1_i386.deb</a>."));
#endif
#ifdef Q_OS_MAC
        emit message(MT_ERROR, tr("Please install mac"));
#endif
        return;
    }
    emit message(MT_INFORMATION, tr("Please wait while Qoobar is splitting the file %1").arg(inputFile));

    QDateTime dt=QDateTime::currentDateTime();

    if (!process) {
        process = new QProcess(this);
        process->disconnect();
        process->setProcessChannelMode(QProcess::MergedChannels);
        process->setReadChannel(QProcess::StandardOutput);

#ifdef Q_OS_MAC
        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        env.insert("PATH", env.value("PATH") + ":" + App->paths->utilitiesPath());
        process->setProcessEnvironment(env);
#endif
    }

    QStringList arguments=QString("split.file.-P.dot.-f.cueFile.-d.outputDir.-O.always.-o.flac").split(QSL("."));
    arguments[1]=inputFile;
    arguments[5]=_cueFile;
    arguments[7]=_outputDir;
#ifdef Q_OS_WIN
    arguments=QString("split.-P.dot.-f.cueFile.-d.outputDir.-O.always.-o.flac.file").split(QSL("."));
    arguments[11]=inputFile;
    arguments[4]=_cueFile;
    arguments[6]=_outputDir;
#endif
#ifdef Q_OS_MAC
    arguments=QString("split.-P.dot.-f.cueFile.-d.outputDir.-O.always.-o.flac.file").split(QSL("."));
    arguments[11]=inputFile;
    arguments[4]=_cueFile;
    arguments[6]=_outputDir;
#endif


    QEventLoop q;
    connect(process,SIGNAL(finished(int)),&q,SLOT(quit()));
    connect(process,SIGNAL(error(QProcess::ProcessError)),&q,SLOT(quit()));
//#ifndef Q_OS_WIN
    connect(process,SIGNAL(readyReadStandardOutput()),SLOT(updateText()));
    connect(process,SIGNAL(readyRead()),SLOT(updateText()));
//#endif
    time = QDateTime::currentDateTime();
    QFileSystemWatcher watcher(this);
    watcher.addPath(_outputDir);
    connect(&watcher,SIGNAL(directoryChanged(QString)),SLOT(outputDirChanged(QString)));

#ifdef Q_OS_MAC
    process->start(App->paths->utilitiesPath()+QSL("/shntool"),arguments);
#else
    process->start(QSL("shntool"),arguments);
#endif
    q.exec();

    int code=process->exitCode();
    if (code==0) {
        QFileInfoList newFiles = QDir(_outputDir).entryInfoList(QStringList()<<QSL("*.flac"),QDir::Files);
        Q_FOREACH(const QFileInfo &newFile,newFiles) {
            if (newFile.created()>dt || newFile.lastModified()>dt)
                files << newFile.canonicalFilePath();
        }
    }


    if (files.size()>0) {
        if (files.at(0).contains(QRegExp(QSL("00\\b")))) {
            //remove pregap file
            QFile::remove(files.at(0));
            files.removeFirst();
        }
    }

    if (!files.isEmpty())
        parseCue();
}

void CueSplitter::setCueFile(const QString &cueFile)
{
     _cueFile = cueFile;
     App->lastCueFile = cueFile;
     _outputDir=QFileInfo(cueFile).canonicalPath();
}

void CueSplitter::parseCue()
{
    QList<Tag> result;
    QFile file(_cueFile);
    QStringList fields;
    if (file.open(QFile::ReadOnly | QFile::Text)) {
        QTextStream in(&file);
        in.setCodec(QTextCodec::codecForLocale());
        in.setAutoDetectUnicode(true);
        while(!in.atEnd()) fields << in.readLine();
    }
    if (fields.isEmpty()) return;

    int currentTrack=-1; //top level corresponds to the whole disk
    QString albumArtist;
    QString comment;
    QString album;
    QString genre;
    QString date;
    for (int i=0; i<fields.count(); ++i) {
        QString s=fields.at(i).simplified();
        if (s.contains(QRegExp(QSL("\\b(INDEX|FILE|PREGAP|POSTGAP|CDTEXTFILE|FLAGS)\\b"))))
            continue;
        if (s.startsWith(QLS("TRACK"))) {
            currentTrack+=1;
            Tag item(QSL(""));
            item.setTag(TRACKNUMBER,QString::number(currentTrack+1));
            item.setTag(ALBUM,album);
            item.setTag(ALBUMARTIST,albumArtist);
            item.setTag(COMMENT,comment);
            item.setTag(YEAR,date);
            item.setTag(GENRE,genre);
            result.append(item);
        }
        else {
            QString ss=s.section(QSL(" "),1);
            if (ss.startsWith(QLS("\""))) ss.remove(0,1);
            if (ss.endsWith(QLS("\""))) ss.chop(1);
            if (s.startsWith(QLS("TITLE"))) {
                if (currentTrack<0) album=ss;
                else result.last().setTag(TITLE,ss);
            }
            else if (s.startsWith(QLS("PERFORMER"))) {
                if (currentTrack<0) albumArtist=ss;
                else result.last().setTag(ARTIST,ss);
            }
            else if (s.startsWith(QLS("REM"))) {
                s=s.mid(4);
                ss=s.section(QSL(" "),1);
                if (ss.startsWith(QLS("\""))) ss.remove(0,1);
                if (ss.endsWith(QLS("\""))) ss.chop(1);
                if (s.startsWith(QLS("GENRE"))) {
                    if (currentTrack<0) genre=ss;
                    else result.last().setTag(GENRE,ss);
                }
                else if (s.startsWith(QLS("DATE"))) {
                    if (currentTrack<0) date=ss;
                    else result.last().setTag(YEAR,ss);
                }
                else if (s.startsWith(QLS("COMMENT"))) {
                    if (currentTrack<0) comment=ss;
                    else result.last().setTag(COMMENT,comment);
                }
            }
        }
    }
    //if all artists is empty, move albumArtist there
    bool isEmpty=true;
    for (int i=0; i<result.size(); ++i)
        if (!result[i].tag(ARTIST).isEmpty()) {
            isEmpty=false;
            break;
        }

    if (isEmpty) {
        for (int i=0; i<result.size(); ++i) {
            result[i].setTag(ARTIST,result[i].tag(ALBUMARTIST));
            result[i].setTag(ALBUMARTIST,QString());
        }
    }
    for (int i=0; i<result.size() && i<files.size(); ++i) {
        result[i].setFile(files.at(i));
        TagsReaderWriter t(&result[i]);
        t.writeTags();
    }
}

void CueSplitter::updateText()
{
    QByteArray b=process->readAll();
    emit textReady(QString::fromUtf8(b.data(),b.size()));
}

void CueSplitter::outputDirChanged(const QString &path)
{
    QFileInfoList newFiles = QDir(path).entryInfoList(QStringList()<<QSL("*.flac"),QDir::Files);
    Q_FOREACH(const QFileInfo &newFile,newFiles) {
        if (newFile.created()>time) {
#ifdef Q_OS_WIN
            emit textReady(tr("Processing %1 ...").arg(newFile.canonicalFilePath()));
#endif
            emit next();
        }
    }
    time = QDateTime::currentDateTime();
}
